import { Point } from './point.js';
import { Rect } from './rect.js';
export var Manipulations;
(function (Manipulations) {
    const RESIZE_SENSITIVITY = 0.1;
    const MOVE_SENSITIVITY = 0.1;
    const EDGE_AFFINITY = 30; // px
    function axialPoint(axis, magnitude) {
        switch (axis) {
            case null: return { x: magnitude, y: magnitude };
            case "x" /* Axis.x */: return { x: magnitude, y: 0 };
            case "y" /* Axis.y */: return { x: 0, y: magnitude };
        }
    }
    function scaleForWorkspace(workArea) {
        return (workArea.x + workArea.y) / 2;
    }
    function wrap(edgeAffinity, fn) {
        return function (rect, workArea) {
            const scale = scaleForWorkspace(workArea);
            const newRect = fn(rect, scale);
            if (edgeAffinity) {
                return Rect.moveWithinAffinity({
                    margin: EDGE_AFFINITY,
                    original: rect,
                    modified: newRect,
                    bounds: workArea,
                });
            }
            else {
                return Rect.scaleWithin(newRect, workArea);
            }
        };
    }
    function resize(direction, axis) {
        return wrap(true, function (rect, scale) {
            const amount = RESIZE_SENSITIVITY * scale * direction;
            const diff = axialPoint(axis, amount);
            // p("scaling diff: " + JSON.stringify(diff) + " on window rect: " + JSON.stringify(rect));
            return {
                // remove half of diff from pos to keep centered
                pos: Point.add(rect.pos, Point.scaleConstant(-0.5, diff)),
                size: Point.add(rect.size, diff)
            };
        });
    }
    Manipulations.resize = resize;
    function move(direction, axis) {
        return wrap(false, function (rect, scale) {
            const amount = MOVE_SENSITIVITY * scale * direction;
            const diff = axialPoint(axis, amount);
            // p("adding diff: " + JSON.stringify(amount) + " to window rect: " + JSON.stringify(rect));
            return {
                pos: Point.add(rect.pos, diff),
                size: rect.size,
            };
        });
    }
    Manipulations.move = move;
    class RectAndArea {
        area;
        rect;
        constructor(rect) {
            this.rect = rect;
            this.area = Rect.area(rect);
        }
        static from(r) {
            return new RectAndArea(r);
        }
        intersection(other) {
            let intersection = Rect.intersection(this.rect, other.rect);
            if (intersection == null)
                return null;
            return new RectAndArea(intersection);
        }
        static descendingSortOrder(a, b) {
            return a.area - b.area;
        }
    }
    function freeRectsAround(workArea, win) {
        let bottomRight = Point.add(win.pos, win.size);
        let top = { pos: Point.ZERO, size: { x: workArea.x, y: win.pos.y } };
        let left = { pos: Point.ZERO, size: { x: win.pos.x, y: workArea.y } };
        let bottom = { pos: { x: 0, y: bottomRight.y }, size: { x: workArea.x, y: workArea.y - bottomRight.y } };
        let right = { pos: { x: bottomRight.x, y: 0 }, size: { x: workArea.x - bottomRight.x, y: workArea.y } };
        return ([top, left, bottom, right]
            .filter(Rect.isPositive)
            .map(RectAndArea.from)
            .sort(RectAndArea.descendingSortOrder));
    }
    function largestFreeRect(workArea, windows, minArea) {
        let freeWindowAreas = windows.map(win => freeRectsAround(workArea, win));
        // log("freeWindowAreas = " + JSON.stringify(freeWindowAreas));
        var largest = null;
        function worthPursuing(candidate) {
            if (candidate == null || candidate.area < minArea)
                return false;
            return largest == null || candidate.area > largest.area;
        }
        // pick one free rect per window
        function recurse(rect, remaining) {
            // log("recursing over rect " + JSON.stringify(rect) + ", with " + remaining.length + " windows remaining");
            if (remaining.length == 0) {
                // we've picked one free rect from each window, and the intersection is nonzero
                if (worthPursuing(rect)) {
                    largest = rect;
                }
                return;
            }
            let nextSet = remaining[0];
            // TODO: we'd save allocations using indexes
            let nextRemaining = remaining.slice(1);
            nextSet.forEach(function (candidate) {
                let intersection = rect.intersection(candidate);
                // log("intersection of " + JSON.stringify(rect) + " and " + JSON.stringify(candidate) + " = " + JSON.stringify(intersection));
                if (worthPursuing(intersection)) {
                    // valid intersection which isn't already smaller than our largest candidate, keep going
                    recurse(intersection, nextRemaining);
                }
            });
        }
        recurse(new RectAndArea({ pos: Point.ZERO, size: workArea }), freeWindowAreas);
        if (largest == null || largest.area < minArea)
            return null;
        return largest.rect;
    }
    Manipulations.largestFreeRect = largestFreeRect;
})(Manipulations || (Manipulations = {}));
