import { assert, p } from "./common.js";
import { Manipulations } from "./manipulations.js";
import { MenuHandlers } from "./menu_handlers.js";
import { MenuSelection } from "./menu_selection.js";
import { Point } from "./point.js";
import { Preview } from "./preview.js";
export var Menu;
(function (Menu_1) {
    function modeForKey(key) {
        switch (key) {
            case 64 /* KeyCode.CTRL */: return 2 /* MouseMode.MOVE */;
            case 133 /* KeyCode.ALT */: return 1 /* MouseMode.RESIZE */;
            case 65 /* KeyCode.SPACE */:
            case 23 /* KeyCode.TAB */: return 3 /* MouseMode.NOOP */;
            default: return null;
        }
    }
    function directionForKey(key) {
        switch (key) {
            case 45 /* KeyCode.K */:
            case 111 /* KeyCode.UP */: return 1 /* Direction.UP */;
            case 44 /* KeyCode.J */:
            case 116 /* KeyCode.DOWN */: return 3 /* Direction.DOWN */;
            case 43 /* KeyCode.H */:
            case 113 /* KeyCode.LEFT */: return 0 /* Direction.LEFT */;
            case 46 /* KeyCode.L */:
            case 114 /* KeyCode.RIGHT */: return 2 /* Direction.RIGHT */;
            default: return null;
        }
    }
    function selectionForKey(key, splitMode) {
        switch (splitMode) {
            case (0 /* SplitMode.FOUR */):
                switch (key) {
                    case 21 /* KeyCode.EQUAL */:
                        return MenuSelection.Inner(2 /* Anchor.TOP */, splitMode);
                    case 20 /* KeyCode.MINUS */:
                        return MenuSelection.Inner(6 /* Anchor.BOTTOM */, splitMode);
                }
                break;
            case (1 /* SplitMode.SIX */):
                switch (key) {
                    case 30 /* KeyCode.U */:
                        return MenuSelection.Inner(1 /* Anchor.TOPLEFT */, splitMode);
                    case 31 /* KeyCode.I */:
                        return MenuSelection.Inner(6 /* Anchor.BOTTOM */, splitMode);
                    case 32 /* KeyCode.O */:
                        return MenuSelection.Inner(3 /* Anchor.TOPRIGHT */, splitMode);
                }
                break;
        }
        return null;
    }
    class Menu {
        ui;
        preview;
        menuHandlers;
        menu;
        mouseMode;
        Sys;
        win;
        splitMode;
        cleanup = [];
        static show(Sys, parent, origin, win) {
            if (!win) {
                p("no active window; ignoring");
                return null;
            }
            return new Menu(Sys, parent, origin, win);
        }
        constructor(Sys, parent, origin, win) {
            const screen = Sys.workspaceArea(win);
            p("getting windowRect of win " + win);
            const windowRect = Sys.windowRect(win);
            assert(windowRect, "windowRect()");
            const screenRatio = (screen.x / Math.max(screen.y, 1));
            p("Screen ratio: " + screenRatio);
            this.splitMode = screenRatio > 2 ? 1 /* SplitMode.SIX */ : 0 /* SplitMode.FOUR */;
            this.Sys = Sys;
            p("creating menu at " + JSON.stringify(origin) + " with bounds " + JSON.stringify(screen) + " and split mode " + this.splitMode);
            const self = this;
            this.win = win;
            this.mouseMode = 0 /* MouseMode.MENU */;
            const modal = Sys.newClutterActor();
            modal.set_size(screen.x, screen.y);
            modal.set_reactive(true);
            const menu = this.menu = Sys.newClutterActor();
            const menuSize = { x: 200, y: 200 };
            menu.set_size(menuSize.x, menuSize.y);
            const canvas = Sys.newClutterCanvas();
            canvas.set_size(menuSize.x, menuSize.y);
            menu.set_content(canvas);
            const position = Point.subtract(origin, Point.scaleConstant(0.5, menuSize));
            menu.set_position(position.x, position.y);
            const preview = this.preview = new Preview.LayoutPreview(Sys, screen, windowRect, win);
            const handlers = this.menuHandlers = new MenuHandlers.Handlers(Sys, self.splitMode, menuSize, origin, canvas, preview);
            canvas.connect('draw', handlers.draw);
            modal.connect('motion-event', function (_actor, event) {
                let position = Sys.translateEventCoordinates(Point.ofEvent(event), win);
                self.menuHandlers.onMouseMove(self.mouseMode, self.splitMode, position);
                return Sys.Clutter.EVENT_PROPAGATE;
            });
            modal.connect('key-press-event', function (_actor, event) {
                self.onKeyPress(event);
                return Sys.Clutter.EVENT_STOP;
            });
            modal.connect('key-release-event', function (_actor, event) {
                self.onKeyRelease(event);
                return Sys.Clutter.EVENT_STOP;
            });
            modal.connect('button-press-event', function () {
                self.complete(true);
                return Sys.Clutter.EVENT_STOP;
            });
            const coverPane = Sys.newClutterActor();
            this.ui = coverPane;
            modal.set_reactive(true);
            modal.add_actor(menu);
            coverPane.add_actor(this.preview.ui);
            coverPane.add_actor(modal);
            parent.insert_child_above(coverPane, null);
            const grab = Sys.pushModal(modal);
            p("pushModal succeeded");
            this.cleanup.push(() => {
                p("Popping modal");
                Sys.popModal(grab);
                parent.remove_child(coverPane);
            });
            modal.grab_key_focus();
            canvas.invalidate();
        }
        onKeyPress(event) {
            const code = event.get_key_code();
            p('keypress: ' + code);
            if (code == 9 /* KeyCode.ESC */) {
                this.complete(false);
            }
            else if (code == 36 /* KeyCode.RETURN */) {
                this.complete(true);
            }
            else if (code == 38 /* KeyCode.A */ && this.mouseMode == 0 /* MouseMode.MENU */) {
                if (this.splitMode == 0 /* SplitMode.FOUR */) {
                    this.splitMode = 1 /* SplitMode.SIX */;
                }
                else {
                    this.splitMode = 0 /* SplitMode.FOUR */;
                }
                this.menuHandlers.updateMenuSelection(this.splitMode);
            }
            else {
                const newMode = modeForKey(code);
                if (newMode !== null) {
                    if (this.mouseMode !== newMode) {
                        if (this.menuHandlers.trackMouse(this.mouseMode)) {
                            p("entering mode " + newMode);
                            this.mouseMode = newMode;
                            // hide the menu, we're showing a window manipulation rect
                            this.menu.hide();
                        }
                        else {
                            p("not entering " + newMode + " due to current menu selection");
                        }
                    }
                }
                else if (this.mouseMode == 0 /* MouseMode.MENU */) {
                    const direction = directionForKey(code);
                    const selection = selectionForKey(code, this.splitMode);
                    if (direction !== null) {
                        this.menuHandlers.applyDirection(direction);
                    }
                    else if (selection !== null) {
                        this.menuHandlers.updateSelection(selection);
                    }
                }
                else {
                    this.onManipulationKeyPress(code, event.get_state());
                }
            }
        }
        onManipulationKeyPress(code, modifiers) {
            // p("keycode: " + code);
            // p("Modifiers: " + modifiers);
            if (modifiers & this.Sys.Clutter.ModifierType.SHIFT_MASK) {
                switch (code) {
                    case 43 /* KeyCode.H */:
                    case 113 /* KeyCode.LEFT */: return this.preview.applyManipulator(Manipulations.resize(-1, "x" /* Axis.x */));
                    case 30 /* KeyCode.U */:
                    case 44 /* KeyCode.J */:
                    case 116 /* KeyCode.DOWN */: return this.preview.applyManipulator(Manipulations.resize(1, "y" /* Axis.y */));
                    case 31 /* KeyCode.I */:
                    case 45 /* KeyCode.K */:
                    case 111 /* KeyCode.UP */: return this.preview.applyManipulator(Manipulations.resize(-1, "y" /* Axis.y */));
                    case 46 /* KeyCode.L */:
                    case 114 /* KeyCode.RIGHT */: return this.preview.applyManipulator(Manipulations.resize(1, "x" /* Axis.x */));
                }
            }
            else {
                switch (code) {
                    case 43 /* KeyCode.H */:
                    case 113 /* KeyCode.LEFT */: return this.preview.applyManipulator(Manipulations.move(-1, "x" /* Axis.x */));
                    case 30 /* KeyCode.U */:
                    case 44 /* KeyCode.J */:
                    case 116 /* KeyCode.DOWN */: return this.preview.applyManipulator(Manipulations.move(1, "y" /* Axis.y */));
                    case 31 /* KeyCode.I */:
                    case 45 /* KeyCode.K */:
                    case 111 /* KeyCode.UP */: return this.preview.applyManipulator(Manipulations.move(-1, "y" /* Axis.y */));
                    case 46 /* KeyCode.L */:
                    case 114 /* KeyCode.RIGHT */: return this.preview.applyManipulator(Manipulations.move(1, "x" /* Axis.x */));
                    case 20 /* KeyCode.MINUS */: return this.preview.applyManipulator(Manipulations.resize(-1, null));
                    case 21 /* KeyCode.EQUAL */: return this.preview.applyManipulator(Manipulations.resize(1, null));
                }
            }
        }
        onKeyRelease(event) {
            const code = event.get_key_code();
            const fromState = modeForKey(code);
            if (fromState != null && this.mouseMode == fromState) {
                p("ending mode " + fromState);
                this.mouseMode = 3 /* MouseMode.NOOP */;
            }
        }
        onSelect(action, rect) {
            switch (action) {
                case 1 /* Action.MINIMIZE */:
                    this.Sys.minimize(this.win);
                    return;
                case 2 /* Action.RESIZE */:
                    this.Sys.moveResize(this.win, rect);
                    break;
                case 0 /* Action.CANCEL */:
                    break;
            }
            this.Sys.activate(this.win);
        }
        destroy() {
            p("hiding menu");
            this.cleanup.forEach(function (fn) {
                try {
                    fn();
                }
                catch (e) {
                    p("Cleanup action failed: " + e);
                }
            });
            this.cleanup = [];
        }
        complete(accept) {
            if (!accept) {
                this.onSelect(0 /* Action.CANCEL */, null);
            }
            else {
                const selection = this.menuHandlers.getSelection();
                if (selection.eqTo(1 /* Ring.INNER */, 6 /* Anchor.BOTTOM */, 0 /* SplitMode.FOUR */)) {
                    this.onSelect(1 /* Action.MINIMIZE */, null);
                }
                else {
                    const rect = this.preview.getRect();
                    if (rect !== null) {
                        this.onSelect(2 /* Action.RESIZE */, rect);
                    }
                }
            }
            this.destroy();
        }
    }
    Menu_1.Menu = Menu;
})(Menu || (Menu = {}));
